#include "binner.h"
#include <map>

bool isclosekmer(const ULL4 &a,const ULL4 &b)
{
	ULL4 r = a^b;
	if(r.high)
	{
		if(r.mid_H | r.mid_L | r.low)
			return false;
		unsigned long long v=((r.high>>1)|r.high)&0x5555555555555555;
		if(v & v-1)
			return false;
		else return true;
	}
	else if(r.mid_H)
	{
		if(r.mid_L | r.low)
			return false;
		unsigned long long v=((r.mid_H>>1)|r.mid_H)&0x5555555555555555;
		if(v & v-1)
			return false;
		else return true;
	}
	else if(r.mid_L)
	{
		if(r.low)
			return false;
		unsigned long long v=((r.mid_L>>1)|r.mid_L)&0x5555555555555555;
		if(v & v-1)
			return false;
		else return true;
	}
	else if(r.low)
	{
		unsigned long long v=((r.low>>1)|r.low)&0x5555555555555555;
		if(v & v-1)
			return false;
		else return true;
	}
	else 
	{
		std::cerr << "ERROR in isclosekmer :\t" << a << "\t" << b;
		return false;
	}
}

int kmerdiffer(const ULL4 &a,const ULL4 &b)
{
	int result = -1;
	ULL4 r = a^b;
	if(r.low)
		result = diffsearch((r.low>>1|r.low)&0x5555555555555555)>>1;
	else if(r.mid_L)
		result = 32 + (diffsearch((r.mid_L>>1|r.mid_L)&0x5555555555555555)>>1);
	else if(r.mid_H)
		result = 64 + (diffsearch((r.mid_H>>1|r.mid_H)&0x5555555555555555)>>1);
	else if(r.high)
		result = 96 + (diffsearch((r.high>>1|r.high)&0x5555555555555555)>>1);
	return result;
}

unsigned diffsearch(unsigned long long a)
{
	unsigned long  pos = 0;
	__asm__("bsrq %1,%0\n\t"
					: "+r" (pos)
					: "rm" (a));
	return pos;
}

string ull2str(unsigned long long kmer, int size)//turn the lower (size*2) bits into DNA string
{
	static char _a[32];
	for(int i=size-1;i>=0;++i)
	{
		_a[i] = kmer & 0x3;
		kmer >>= 2;
	}
	_a[size]=0;
	return string(_a);
}

string read2str(const ULL4& read,int ReadLen)
{
	if(ReadLen > 96)
		return (ull2str(read.high,ReadLen - 96) + ull2str(read.mid_H,32) + ull2str(read.mid_L,32) + ull2str(read.low,32));
	else if(ReadLen > 64)
		return (ull2str(read.mid_H,ReadLen - 64) + ull2str(read.mid_L,32) + ull2str(read.low,32));
	else if(ReadLen > 32)
		return (ull2str(read.mid_L,ReadLen - 32) + ull2str(read.low,32));
	else
		return (ull2str(read.low,ReadLen));
}

int ullDist(unsigned long long a, unsigned long long b)
{
	int result = 0;
	unsigned long long c = a^b;
	c = ((c>>1)|c)&0x5555555555555555;
	for(;c;++result)
	{
		c &= c-1;
	}
	return result;
}

//we assume a[i] is 0~3, so we don't check it here.
ULL4 array2read(int a[], int readLen)
{
	int i=0;
	unsigned long long high=0ULL,mid_H=0ULL,mid_L=0ULL,low=0ULL;
	for(;i<readLen-96;++i)
	{
		high <<= 2;
		high |= a[i];
	}
	for(;i<readLen-64;++i)
	{
		mid_H <<= 2;
		mid_H |= a[i];
	}
	for(;i<readLen-32;++i)
	{
		mid_L <<= 2;
		mid_L |= a[i];
	}
	for(;i<readLen;++i)
	{
		low <<= 2;
		low |= a[i];
	}
	return ULL4(high,mid_H,mid_L,low);
}

//we assume a[i] is 0~3, so we don't check it here.
//rc means reverse complement
ULL4 array2rcread(int a[], int readLen)
{
	int i=readLen-1;
	unsigned long long high=0ULL,mid_H=0ULL,mid_L=0ULL,low=0ULL;
	for(;i>=96;--i)
	{
		high <<= 2;
		high |= 3-a[i];
	}
	for(;i>=64;--i)
	{
		mid_H <<= 2;
		mid_H |= 3-a[i];
	}
	for(;i>=32;--i)
	{
		mid_L <<= 2;
		mid_L |= 3-a[i];
	}
	for(;i>=0;--i)
	{
		low <<= 2;
		low |= 3-a[i];
	}
	return ULL4(high,mid_H,mid_L,low);
}

void compa_read(int ReadLen,const int &position1,const int &position2,int &match,int &mismatch,const ULL4 &read1,const ULL4 &read2)
{
	if(position1 > position2)
	{
		int k=position1-position2;
		ULL4 differ= (read2>>(2*k));
		differ ^= (read1 & AndMask[ReadLen-k]);
		mismatch = differ.non0base();
		match = ReadLen-k-mismatch;
		//Note: can change == to ^, so that can return a value;
	}
	else if(position1 < position2)
	{
		int k=position2-position1;
		ULL4 differ= (read1>>(2*k));
		differ ^= (read2 & AndMask[ReadLen-k]);
		mismatch = differ.non0base();
		match = ReadLen-k-mismatch;
		//Note: can change == to ^, so that can return a value;
	}
	else
	{
		ULL4 differ= read1;
		differ ^= read2;
		mismatch = differ.non0base();
		match = ReadLen-mismatch;
	}
}
//////////////////////////////////////////////////////////////////////////////////
////Phase 1
void USMerge(const vector<pairlong>& ITR, USet* Type, int LARGESIZE, int ReadLen,
	   	ULL4*  Reads, ULL4*  RevReads, int StrLen, int StrLenLowCover,bool*IsOrphan)
{
	omp_lock_t type_lock;
	omp_init_lock(&type_lock);
	int itrstart = 0;
	const unsigned itrend = ITR.size();
	const int Frag = 100;
	unsigned ParaThr = 0;
	while(ParaThr < itrend && INodePool.getVSize(ITR[ParaThr].node) < 10000)
		++ParaThr;
//////////////////////////////////////////////////////////
	map<int,int> TableCount;
	for(int i=0;i<itrend;++i)
		++TableCount[INodePool.getVSize(ITR[i].node)/10];
	std::cout << "Table Count:\t" << std::endl;
	for(map<int,int>::const_iterator itr = TableCount.begin();itr!=TableCount.end();++itr)
		std::cout << itr->first << "\t" << itr->second << std::endl;
//	while(itrend < ITR.size() && INodePool.getVSize(ITR[itrend].node) == INodePool.getVSize(ITR[itrstart].node))
	unsigned ITRend  = ITR.size();
	cerr<<"ParaThr:\t"<<ParaThr << ' ' << itrend << endl;
	//std::cerr<<"Frag:\t"<<Frag<<endl;
//	while(itrend < ITR.size())
	unsigned ParaThr2 = ParaThr;
	while(ParaThr2 < itrend && INodePool.getVSize(ITR[ParaThr2].node) < 100000)
		++ParaThr2;
#pragma omp parallel for schedule(dynamic)
		for(unsigned itrind = 0;itrind<ParaThr;++itrind)
		{
			IntNode* node = INodePool.getRef(ITR[itrind].node);
			int sub_size = (node->VSize);
			if(sub_size<2)
			{
				continue;
			}
//			cerr << 'a' << "\t" << sub_size << endl;
	
			int* indexes=new int[sub_size];
			int* positions=new int[sub_size];int temp=0;
			int* isRev=new int[sub_size];
			for(NodeIDPosi* map_it2 = node->myvector;map_it2 != (node->myvector)+node->VSize; ++map_it2)
			{
	//`			if(IsOrphan[(*map_it2).id])continue;
				indexes[temp]=((*map_it2).id);
				isRev[temp]=((*map_it2).posi & 1U);//last bit is a flag
				positions[temp++]=((*map_it2).posi)>>1;//last bit is a flag
			}
	//		sub_size = temp;
//			cerr << 'b' << endl;
	
			ULL4 read1,read2;
			for(int i=0;i<sub_size;++i)
			{
				int indexes1 = indexes[i];
	
				if(isRev[i])	read1 = RevReads[indexes1];
				else	read1 = Reads[indexes1];
	
//				cerr << 'c' << endl;
				int j=i+1;
				while(j<sub_size && positions[j]-positions[i]<=ReadLen-StrLen)
					++j;
				for(;j<sub_size;++j)
				{
					int indexes2 = indexes[j];
					if(Type->find(indexes1)==Type->find(indexes2))
						continue;
					if(positions[j]-positions[i]>ReadLen-StrLenLowCover)
						break;
//					cerr << 'd' << endl;
					/////////////////////////////////////////////////////////////////////////////////////
					if((Type->getsize(indexes1)<LARGESIZE && Type->getsize(indexes2)<LARGESIZE )||(Type->getsize(indexes1)<Frag) || Type->getsize(indexes2)<Frag)
					{
						if(isRev[j])	read2 = RevReads[indexes2];
						else read2 = Reads[indexes2];
						int match,mismatch;
						compa_read(ReadLen, positions[i],positions[j],match,mismatch,read1,read2);
					/////////////////bin small ones
						if(match-StrLenLowCover>=0 && (match-StrLenLowCover >= confidence[match+mismatch-StrLenLowCover]))
						{
							omp_set_lock(&type_lock);
							Type->Union(indexes1,indexes2);
							omp_unset_lock(&type_lock);
						}
					}
	//				cerr << 'e' << endl;
				}
			}
			delete[]indexes;delete[]positions;delete[]isRev;
		}
		cerr << " To ParaThr."<<endl;
		cout << " To ParaThr."<<endl;
		system("date");

#pragma omp parallel for schedule(dynamic)
		for(unsigned itrind = ParaThr;itrind<ParaThr2;++itrind)
//		for(unsigned itrind = ParaThr;itrind<itrend;++itrind)
		{
			IntNode* node = INodePool.getRef(ITR[itrind].node);
			int sub_size = (node->VSize);
			if(sub_size<2)
			{
				continue;
			}
	
			int* indexes=new int[sub_size];
			int* positions=new int[sub_size];int temp=0;
			int* isRev=new int[sub_size];
			for(NodeIDPosi* map_it2 = node->myvector;map_it2 != (node->myvector)+node->VSize; ++map_it2)
			{
	//			if(IsOrphan[(*map_it2).id])continue;
				indexes[temp]=((*map_it2).id);
				isRev[temp]=((*map_it2).posi & 1U);//last bit is a flag
				positions[temp++]=((*map_it2).posi)>>1;//last bit is a flag
			}
	//		sub_size = temp;
#pragma omp parallel for schedule(dynamic)
			for(int i=0;i<sub_size;++i)
			{
				ULL4 read1,read2;
				int indexes1 = indexes[i];
	
				if(isRev[i])	read1 = RevReads[indexes1];
				else	read1 = Reads[indexes1];
	
				int j=i+1;
				while(j<sub_size && positions[j]-positions[i]<=ReadLen-StrLen)
					++j;
				for(;j<sub_size;++j)
				{
					int indexes2 = indexes[j];
					if(Type->find(indexes1)==Type->find(indexes2))
						continue;
					if(positions[j]-positions[i]>ReadLen-StrLenLowCover)
						break;
					/////////////////////////////////////////////////////////////////////////////////////
					if((Type->getsize(indexes1)<LARGESIZE && Type->getsize(indexes2)<LARGESIZE )||(Type->getsize(indexes1)<Frag) || Type->getsize(indexes2)<Frag)
					{
						if(isRev[j])	read2 = RevReads[indexes2];
						else read2 = Reads[indexes2];
						int match,mismatch;
						compa_read(ReadLen, positions[i],positions[j],match,mismatch,read1,read2);
					/////////////////bin small ones
						if(match-StrLenLowCover>=0 && (match-StrLenLowCover >= confidence[match+mismatch-StrLenLowCover]))
						{
							omp_set_lock(&type_lock);
							Type->Union(indexes1,indexes2);
							omp_unset_lock(&type_lock);
						}
					}
				}
			}
			delete[]indexes;delete[]positions;delete[]isRev;
		}

	omp_destroy_lock(&type_lock);
}

/////////////////////////////////////////////////////////////
/////////////////////////////////////////////////////////////
void iniMap()
{
#pragma omp parallel for
	for(int i=0;i<HASHSIZE;++i)
		KmerMap[i] = INULL;
}

inline void increaseHash(KmerType id)
{
	unsigned hashid = id & HASHMASK;
	INodeRef curr = KmerMap[hashid];
	if(curr == INULL)
	{
		KmerMap[hashid] = INodePool.getNew(id,1);
		++NodeCount;
		return;
	}
	INodeRef prev = curr;
	for(;curr != INULL;curr = INodePool.getNext(curr))
	{
		if(INodePool.getKmer(curr)==id)break;
		prev = curr;
	}
	if(curr == INULL)
	{
		INodePool.setNext(prev,INodePool.getNew(id,1));
		++NodeCount;
	}
	else if(INodePool.getKmer(curr) == id)
		INodePool.incVSize(curr);
}

inline void shrinkStep()
{
#pragma omp parallel for
	for(int i=0;i<HASHSIZE;++i)
	{
		//////////////////////////////////////remove nodes with vector size 1
		INodeRef prev = KmerMap[i];
		if(prev == INULL)continue;
		INodeRef curr = INodePool.getNext(prev);
		while(curr!=INULL)
		{
			if(INodePool.getVSize(curr)<=1)
			{
				INodePool.setNext(prev,INodePool.getNext(curr));
				INodePool.setVSize(curr,0);
				curr = INodePool.getNext(prev);
			}
			else
			{
				prev = curr;
				curr = INodePool.getNext(prev);
			}
		}
		curr = KmerMap[i];
		if(INodePool.getVSize(curr)<=1)
		{
			KmerMap[i] = INodePool.getNext(curr);
			INodePool.setVSize(curr,0);
		}
		//////////////////////test
		for(INodeRef curr = KmerMap[i];curr != INULL;curr = INodePool.getNext(curr))
		{
			if(INodePool.getVSize(curr)<=1)
				cerr<< " Error in iniArrayMap. " << INodePool.getVSize(curr) << endl;
		}
		///////////////////////end test
	}
	INodePool.shrinkSize1(KmerMap);
	std::cout<<"after deletion in shrinkStep." << std::endl;
	std::cerr<<"after deletion in shrinkStep." << std::endl;
	system("ps ux ");
}

void countKmer(const int ReadNum, const int ReadLen, const ULL4* Reads, const ULL4* RevReads)
{
	omp_lock_t* hash_lock = new omp_lock_t[1U<<28];
	for(int i=0;i<(1U<<28);++i)
		omp_init_lock(&hash_lock[i]);
	const unsigned Round = 0x3;
	for(unsigned maskt = 0;maskt <= Round;++maskt)
	{
#pragma omp parallel for schedule(dynamic)
		for(int index = 0;index < ReadNum; ++index)
		{
			for(unsigned int i=0;i<=ReadLen-32;++i)
			{
				KmerType left =   ((Reads[index]>> ((ReadLen-16-i)*2)).low);
				KmerType right=((RevReads[index]>> (i*2)).low);
				if(left <= right)
				{
					if(((left>>28)&Round) == maskt)
					{
						omp_set_lock(&hash_lock[left&0xfffffff]);
						increaseHash(left);
						omp_unset_lock(&hash_lock[left&0xfffffff]);
					}
				}
				else
				{
					if(((right>>28)&Round) == maskt)
					{
						omp_set_lock(&hash_lock[right&0xfffffff]);
						increaseHash(right);
						omp_unset_lock(&hash_lock[right&0xfffffff]);
					}
				}
			}
		}
		shrinkStep();
	}
	for(int i=0;i<(1U<<28);++i)
		omp_destroy_lock(&hash_lock[i]);
	delete[] hash_lock;
}

/*
void countKmer(const int ReadNum, const int ReadLen, const ULL4* Reads, const ULL4* RevReads)
{
	for(int index = 0;index < ReadNum; ++index)
	{
		for(unsigned int i=0;i<=ReadLen-32;++i)
		{
			KmerType left =   (Reads[index]>> ((ReadLen-16-i)*2)).low;
			KmerType right=(RevReads[index]>> (i*2)).low;
			if(left <= right)
				increaseHash(left);
			else
				increaseHash(right);
		}
	}
}
*/

void iniArrayMap()
{
	shrinkStep();
	/*
#pragma omp parallel for
	for(int i=0;i<(1U<<28);++i)
	{
		//////////////////////////////////////remove nodes with vector size 1
		INodeRef prev = KmerMap[i];
		if(prev == INULL)continue;
		INodeRef curr = INodePool.getNext(prev);
		while(curr!=INULL)
		{
			if(INodePool.getVSize(curr)<=1)
			{
				INodePool.setNext(prev,INodePool.getNext(curr));
				INodePool.setVSize(curr,0);
				curr = INodePool.getNext(prev);
			}
			else
			{
				prev = curr;
				curr = INodePool.getNext(prev);
			}
		}
		curr = KmerMap[i];
		if(INodePool.getVSize(curr)<=1)
		{
			KmerMap[i] = INodePool.getNext(curr);
			INodePool.setVSize(curr,0);
		}
		for(INodeRef curr = KmerMap[i];curr != INULL;curr = INodePool.getNext(curr))
		{
			if(INodePool.getVSize(curr)<=1)
				cerr<< " Error in iniArrayMap. " << INodePool.getVSize(curr) << endl;
		}
	}
	INodePool.shrinkSize1(KmerMap);
	*/
	std::cout<<"after deletion." << std::endl;
	std::cerr<<"after deletion." << std::endl;
	system("ps ux ");
//	int vcount = 0;
#pragma omp parallel for schedule(dynamic)
	for(int i=0;i<HASHSIZE;++i)
	{
		//////////////////////////////////////initial nodes
		for(INodeRef curr = KmerMap[i];curr != INULL;curr = INodePool.getNext(curr))
		{
			if(INodePool.getVSize(curr)>1)
			{
				INodePool.newVector(curr);
////////////////////////////////
/*				INodePool.sycV(curr);
#pragma omp atomic
				vcount += INodePool.getVSize(curr);
				*/
////////////////////////////////
				INodePool.setVSize(curr,0);
			}
			else
				cerr << "ERROR! Vector size <= 1\t" << INodePool.getVSize(curr) << endl;
		}
	}
}

inline void insertHash(KmerType id, unsigned ele,unsigned posi)
{
	unsigned hashid = id & HASHMASK;
	for(INodeRef curr = KmerMap[hashid];curr != INULL;curr = INodePool.getNext(curr))
	{
		if(INodePool.getKmer(curr) == id)
		{
			INodePool.push_back(curr,ele,posi);
			return;
		}
	}
}

void mapReads(const int ReadNum, const int ReadLen, const ULL4* Reads, const ULL4* RevReads,bool* IsOrphan)
{
	omp_lock_t* hash_lock = new omp_lock_t[1U<<20];
	for(int i=0;i<(1U<<20);++i)
		omp_init_lock(&hash_lock[i]);
#pragma omp parallel for schedule(dynamic)
	for(int index = 0;index < ReadNum; ++index)
	{
		if(IsOrphan[index])continue;
		for(unsigned int i=0;i<=ReadLen-32;++i)
		{
			KmerType left =   ((Reads[index]>> ((ReadLen-16-i)*2)).low);
			KmerType right=((RevReads[index]>> (i*2)).low);
			///////////////////////////////////////////pick the smaller one
			if(left <= right)
			{
				omp_set_lock(&hash_lock[left&0xfffff]);
				insertHash(left,(index),(i<<1));
				omp_unset_lock(&hash_lock[left&0xfffff]);
			}
			else
			{
				omp_set_lock(&hash_lock[right&0xfffff]);
				insertHash(right,(index),((ReadLen-16-i)<<1)|1U);
				omp_unset_lock(&hash_lock[right&0xfffff]);
			}
		}
	}
	for(int i=0;i<(1U<<20);++i)
		omp_destroy_lock(&hash_lock[i]);
	delete[] hash_lock;
}

void testVector(const vector<pairlong>& ITR)
{
	const unsigned INodeNum = INodePool.getNodeNum();
	int hashsize[18];for(int i=0;i<18;++i)hashsize[i]=0;
	unsigned kmerCount = 0;
	for(int hashid=0;hashid<HASHSIZE;++hashid)
	{
		int tmpcount = 0;
		for(INodeRef curr = KmerMap[hashid];curr != INULL;curr = INodePool.getNext(curr))
		{
			++tmpcount;
			if(curr > INodeNum)
				cerr<<"curr > INodeNum!"<<endl;
		}
		if(tmpcount<17)++hashsize[tmpcount];
		else ++hashsize[17];
		//////////////////////////////////////test nodes with vector size 1
		for(INodeRef curr = KmerMap[hashid];curr != INULL;curr = INodePool.getNext(curr))
		{
			if(INodePool.getVSize(curr)<=1)
				cerr<<"Error1 in testVector! VSize <= 1" << INodePool.getVSize(curr) << endl;
/*			if(INodePool.getVSize(curr) != INodePool.getVCount(curr))
				cerr<<"Error! VSize != tmpcount!" << INodePool.getVSize(curr) << "\t" << INodePool.getVSize(curr) << endl;
			*/
		}
		kmerCount += tmpcount;
	}
	cerr << "INodeNum&HashId:\t"<<INodePool.getNodeNum()<<"\t"<<kmerCount<<endl;

	////////////////////////////////////////////end of remove 
	cout<<"hash size: "<<endl;
	for(int i=0;i<18;++i)cout<<hashsize[i]<<endl;
	for(int i=0;i<18;++i)cerr<<hashsize[i]<<endl;
	cout<<"endl of hash size: "<<endl;
}

void rmSameTypeNodes(USet* Type)
{
	/*
	omp_lock_t type_lock;
	omp_init_lock(&type_lock);
	*/
	const unsigned INodeNum = INodePool.getNodeNum();
#pragma omp parallel for schedule(dynamic)
	for(unsigned node = 0;node<INodeNum;++node)
	{
			int sub_size = INodePool.getVSize(node);
			if(sub_size<2)
			{
				continue;
			}
			int* indexes=new int[sub_size];
			int sizeCount = 0;
			for(NodeIDPosi* map_it2 = INodePool.getVector(node);map_it2 != INodePool.getVector(node)+INodePool.getVSize(node); ++map_it2)
				indexes[sizeCount++]=((*map_it2).id);
			int type0 = Type->find(indexes[0]);
			bool isSame = true;
			for(int i=1;i<sub_size;++i)
			{
				if(Type->find(indexes[i]) != type0) 
				{
					isSame = false;
					break;
				}
			}
			delete[]indexes;
			if(isSame)
			{
				SNodePool.getNew(INodePool.getKmer(node), sub_size, type0);
				INodePool.clrVect(node);
				INodePool.setVSize(node,0);
			}
	}
	INodePool.shrinkSize1(NULL);
}

void testVCount()
{
	const int lastN = 256;//output largest N tablecount.
	const unsigned itrend = INodePool.getNodeNum();
	map<int,int> TableCount;
	for(int i=0;i<itrend;++i)
		++TableCount[(INodePool.getVSize(i))/10];
	std::cout << "Table Count:\t" << itrend << std::endl;
	for(map<int,int>::const_iterator itr = TableCount.begin();itr!=TableCount.end();++itr)
		std::cout << itr->first << "\t" << itr->second << std::endl;

	int tsum = 0;
	int thresh = 0;
	for(map<int,int>::reverse_iterator itr = TableCount.rbegin();itr!=TableCount.rend();++itr)
	{
		tsum += itr->second;
		if(tsum >= lastN)
		{
			thresh = itr->first;
			break;
		}
	}
	std::cout <<"LastN thresh:\t" << thresh << std::endl;
	for(int i=0;i<itrend;++i)
	{
		if(INodePool.getVSize(i)/10 >= thresh)
			std::cout << "Large Size:\t" << hex << INodePool.getKmer(i) << dec << "\t" << INodePool.getVSize(i) << std::endl;
	}
	/*
	for(unsigned itrind = ITR.size()-100;itrind < ITR.size();++itrind)
	{
		IntNode* node = INodePool.getRef(ITR[itrind].node);
		std::cout << "Large ITR Size:\t" <<hex << node->kmer << dec << "\t" << node->VSize << endl;
	}
	*/
}
/*
	for(int hashid=0;hashid<HASHSIZE;++hashid)
	{
		for(INodeRef node = KmerMap[hashid];node != INULL;node = INodePool.getNext(node))
		{
			int sub_size = INodePool.getVSize(node);
			if(sub_size<2)
			{
				continue;
			}
			int* indexes=new int[sub_size];
			int sizeCount = 0;
			for(NodeIDPosi* map_it2 = INodePool.getVector(node);map_it2 != INodePool.getVector(node)+INodePool.getVSize(node); ++map_it2)
				indexes[sizeCount++]=((*map_it2).id);
			int type0 = Type->find(indexes[0]);
			bool isSame = true;
			for(int i=1;i<sub_size;++i)
			{
				if(Type->find(indexes[i]) != type0) 
				{
					isSame = false;
					break;
				}
			}
			delete[]indexes;
			if(isSame)
			{
	//			omp_set_lock(&type_lock);
				SNodePool.getNew(INodePool.getKmer(node), sub_size, type0);
	//			omp_unset_lock(&type_lock);
				///////////////////////////////////////////////////////
				INodePool.clrVect(node);
				INodePool.setVSize(node,0);
			}
		}
		//////////////////////////rm nodes.
		INodeRef prev = KmerMap[hashid];
		if(prev == INULL)continue;
		INodeRef curr = INodePool.getNext(prev);
		while(curr!=INULL)
		{
			if(INodePool.getVSize(curr)<=1)
			{
				INodePool.setNext(prev,INodePool.getNext(curr));
				INodePool.setVSize(curr,0);
				curr = INodePool.getNext(prev);
			}
			else
			{
				prev = curr;
				curr = INodePool.getNext(prev);
			}
		}
		curr = KmerMap[hashid];
		if(INodePool.getVSize(curr)<=1)
		{
			KmerMap[hashid] = INodePool.getNext(curr);
			INodePool.setVSize(curr,0);
		}
	}
	INodePool.shrinkSize1(KmerMap);
//	omp_destroy_lock(&type_lock);
}
*/
